;;;; crunch - primitives

(module (crunch compiler primitives) ()

  (import (scheme base)
  		  (chicken base))
  (import matchable srfi-1)

  (import (crunch declarations)
          (crunch compiler type)
          (crunch compiler node))

  (define (promote t1 t2)
    (if (eq? t1 t2)
        t1
        (match (list t1 t2)
          (('integer 'float) 'float)
          (('float 'integer) 'float)
          (('complex (or 'integer 'float)) 'complex)
          (((or 'integer 'float) 'complex) 'complex)
          (_ #f))))

  (define (binary-numeric-predicate args)
    (match args
      (('integer 'integer) '((integer integer) boolean))
      (('float (or 'integer 'float)) '((float float) boolean))
      (((or 'integer 'float) 'float) '((float float) boolean))
      (((or 'integer 'float 'complex) 'complex) '((complex complex) boolean))
      (('complex (or 'integer 'float 'complex)) '((complex complex) boolean))
      (('complex (or 'integer 'float 'complex)) '((complex complex) boolean))
      (_ #f)))

  (define (binary-arithmetic args)
    (match args
      (((and t (or 'integer 'float 'complex))) `((,t) ,t))
      (('integer 'integer) '((integer integer) integer))
      (('float (or 'integer 'float)) '((float float) float))
      (((or 'integer 'float) 'float) '((float float) float))
      (((or 'integer 'float 'complex) 'complex) '((complex complex) complex))
      (('complex (or 'integer 'float 'complex)) '((complex complex) complex))
      (('complex (or 'integer 'float 'complex)) '((complex complex) complex))
      (_ #f)))

  (define (numeric-predicate args)
    (match args
      (((and t (or 'integer 'float 'complex))) `((,t) boolean))
      (_ #f)))

  (define (inexact-function args)
    (match args
      (((and t (or 'integer 'float 'complex))) `((,t) float))
      (_ #f)))

  (define (numeric-function args)
    (match args
      (((and t (or 'integer 'float 'complex))) `((,t) ,t))
      (_ #f)))

  ;; (scheme base)

  (define-overloaded-primitive (scheme#zero? ?) boolean $zero?
    (args (numeric-predicate args)))

  (define-specialization $zero? (_ '(operator == 0)))

  (define-specialization $eq? (_ '(operator ==)))

  (define-specialization $eqv? (_ '(operator ==)))

  (define-overloaded-primitive (scheme#+ ? #!optional ?) ? $+
    (args (binary-arithmetic args)))

  (define-specialization $+
    ((_ _) '(operator +))
    ((_) '(operator + 0)))

  (define-overloaded-primitive (scheme#- ? #!optional ?) ? $-
    (args (binary-arithmetic args)))

  (define-specialization $-
    ((_ _) '(operator -))
    ((_) '(operator * -1)))

  (define-overloaded-primitive (scheme#* ? #!optional ?) ? $*
    (args (binary-arithmetic args)))

  (define-specialization $*
    ((_ _) '(operator *))
    ((_) '(operator * 1)))

  (define-overloaded-primitive (scheme#/ ? #!optional ?) ? $/
    (args (binary-arithmetic args)))

  (define-specialization $/
    ((_ _) '(operator /))
    ((_) 'crunch_inverse))

  (define-primitive (scheme#not boolean) boolean $not)

  (define-specialization $not (_ '(operator !)))

  (define-overloaded-primitive (scheme#= ? ?) boolean $eq?
    (args (binary-numeric-predicate args)))

  (define-overloaded-primitive (scheme#< ? ?) boolean $<
    (args (binary-numeric-predicate args)))

  (define-specialization $< (_ '(operator <)))

  (define-overloaded-primitive (scheme#> ? ?) boolean $>
    (args (binary-numeric-predicate args)))

  (define-specialization $> (_ '(operator >)))

  (define-overloaded-primitive (scheme#<= ? ?) boolean $<=
    (args (binary-numeric-predicate args)))

  (define-specialization $<= (_ '(operator <=)))

  (define-overloaded-primitive (scheme#>= ? ?) boolean $>=
    (args (binary-numeric-predicate args)))

  (define-specialization $>= (_ '(operator >=)))

  (define-primitive (scheme#newline #!optional port) void crunch_newline)

  (define-primitive (scheme#string-length string) integer crunch_string_length)

  (define-primitive (scheme#string #!optional char char char char char char char char char char)
    string crunch_cons_string)

  (define-overloaded-primitive (scheme#vector #!optional ? ? ? ? ? ? ? ? ? ?) ?
    $vector
    ((t more ...)
      (and (not (eq? t '?))
           (let loop ((t t) (m more))
             (cond ((null? m)
             	     `(,(make-list (+ 1 (length more)) t) (vector ,t)))
             	   ((equal? t (car m))
             	     (loop t (cdr m)))
             	   ((promote t (car m)) =>
             	     (lambda (t2)
             	       (loop t2 (cdr m))))
             	   (else #f))))))

  (define-specialization $vector
    (('char . _) 'crunch_cons_u32vector)
    (('integer . _) 'crunch_cons_intvector)
    (('float . _) 'crunch_cons_f64vector)
    (('complex . _) 'crunch_cons_c128vector)
    (('boolean . _) 'crunch_cons_u8vector)
    ((t . _) 'crunch_cons_vector))

  (define-overloaded-primitive (scheme#vector-length ?) integer $vector-length
    ((('vector t)) `(((vector ,t)) integer))
    ((('numvector n)) `(((numvector ,n)) integer)))

  (define-specialization $vector-length
    ((((or 'vector 'numvector) _)) 'crunch_vector_length))

  (define-primitive (scheme#string=? string string) boolean crunch_string_equal)

  (define-primitive (scheme#string>? string string) boolean crunch_string_greater)

  (define-primitive (scheme#string<? string string) boolean crunch_string_less)

  (define-primitive (scheme#string>=? string string) boolean crunch_string_greater_or_equal)

  (define-primitive (scheme#string<=? string string) boolean crunch_string_less_or_equal)

  (define-primitive (scheme#make-string integer #!optional char)
    string crunch_make_string)

  (define-overloaded-primitive (scheme#positive? ?) boolean $positive?
    (((and t (or 'integer 'float 'complex))) `((,t) boolean)))

  (define-specialization $positive? (_ '(operator > 0)))

  (define-overloaded-primitive (scheme#negative? ?) boolean $negative?
    (((and t (or 'integer 'float 'complex))) `((,t) boolean)))

  (define-specialization $negative? (_ '(operator < 0)))

  (define-overloaded-primitive (scheme#eof-object? ?) boolean $eof-object?
    (((and t (or ('vector _) '(numvector u8) 'integer 'char))) `((,t) boolean)))

  (define-specialization $eof-object?
    (((or '(numvector u8) ('vector _))) 'crunch_eofp)
    (((or 'char 'integer)) '(operator == crunch_eof_scalar)))

  (define-overloaded-primitive (scheme#eq? ? ?) boolean $eq?
    ((t1 t2)
      (and (not (eq? t1 '?))
           (not (eq? t2 '?))
           `((,t1 ,t2) boolean))))

  (define-overloaded-primitive (scheme#eqv? ? ?) boolean $eqv?
    (args
      (or (binary-numeric-predicate args)
          (and (memq '? args)
               `(,args boolean)))))

  (define-overloaded-primitive (##sys#eqv? ? ?) boolean $eqv?
    (args
      (or (binary-numeric-predicate args)
          (and (memq '? args)
               `(,args boolean)))))

  (define-overloaded-primitive (scheme#equal? ? ?) boolean $equal?
    ((t1 t2)
      (and (not (eq? t1 '?))
           (not (eq? t2 '?))
           `((,t1 ,t2) boolean))))

  (define-specialization $equal?
    (((and t1 (or ('vector _) 'string ('numvector _) 'port)) t2)
      (if (equal? t1 t2)
          'crunch_equalp
          '(operator ==)))
    (_ '(operator ==)))

  (define-overloaded-primitive (scheme#abs ?) ? $abs
    (args (numeric-function args)))

  (define-specialization $abs
    (((or 'integer 'float)) 'abs)
    (('complex) 'cabs))

  (define-overloaded-primitive (scheme#expt ? ?) ? $expt
    (('integer 'integer) '((integer integer) integer))
    (('float 'float) '((float float) float))
    (('complex 'complex) '((complex complex) complex)))

  (define-specialization $expt
    (('integer 'integer) 'crunch_integer_expt)
    (('float 'float) 'pow)
    (('complex 'complex) 'cpow))

  (define-primitive (scheme#quotient integer integer) integer $quotient)

  (define-specialization $quotient (_ '(operator /)))

  (define-primitive (scheme#remainder integer integer) integer $remainder)

  (define-specialization $remainder (_ 'crunch_remainder))

  (define-primitive (scheme#modulo integer integer) integer $modulo)

  (define-specialization $modulo (_ 'crunch_modulo))

  (define-overloaded-primitive (scheme#square ?) ? $square
    (((and t (or 'integer 'float 'complex))) `((,t) ,t)))

  (define-specialization $square (_ 'crunch_square))

  (define-inline-primitive (scheme#dynamic-wind t1 t2 t3)
    (begin
      (t1)
      (let ((r (t2)))
        (t3)
        r)))

  (define-inline-primitive (scheme#values x) x)

  (define-inline-primitive (##sys#values x) x)

  (define-inline-primitive (scheme#call-with-values p1 p2)
    (p2 (p1)))

  (define-inline-primitive (##sys#call-with-values p1 p2)
    (p2 (p1)))

  (define-overloaded-primitive (scheme#vector-ref ? integer) ? $vector-ref
    ((('vector t) 'integer) `(((vector ,t) integer) ,t)))

  (define-specialization $vector-ref
    ((('vector (or 'integer 'float 'complex 'boolean 'char)) integer) 'crunch_numvector_ref)
    (_ 'crunch_vector_ref))

  (define-overloaded-primitive (scheme#vector-set! ? integer ?)
    void $vector-set!
    ((('vector t) 'integer t2)
      (and (equal? t t2)
           `(((vector ,t) integer ,t) void))))

  (define-specialization $vector-set!
    ((('vector (or 'boolean 'char 'integer 'float 'complex)) integer _) 'crunch_numvector_set)
    (_ 'crunch_vector_set))

  (define-inline-primitive (scheme#vector-for-each proc vec)
    (let ((loop (unspecified)))
      (begin
       (set! loop
         (lambda (vec proc i len)
           (if (scheme#>= i len)
               (unspecified)
               (begin
                 (proc (scheme#vector-ref vec i))
                 (loop vec proc (scheme#+ i 1) len)))))
       (loop vec proc 0 (scheme#vector-length vec)))))

  (define-inline-primitive (scheme#vector-map proc vec)
    (let ((loop (unspecified))
          (nvec (unspecified))
          (vec1 vec))
      (begin
        (set! nvec (scheme#vector-copy vec1))
        (set! loop
          (lambda (vec proc i len)
            (if (scheme#>= i len)
                nvec
                (begin
                  (scheme#vector-set! nvec i (proc (scheme#vector-ref vec i)))
                  (loop vec proc (scheme#+ i 1) len)))))
        (loop vec1 proc 0 (scheme#vector-length vec)))))

  (define-overloaded-primitive (scheme#vector-copy ? #!optional integer integer)
    ? $vector-copy
    (((and t (or ('vector _) ('numvector _))) . more)
      `((,t ,@more) ,t)))

  (define-specialization $vector-copy
    ((_) 'crunch_vector_copy1)
    (_ 'crunch_vector_copy))

  (define-overloaded-primitive (scheme#vector-copy! ? integer ? integer
    #!optional integer integer) void $vector-copy!
    ((('vector t) at ('vector t2) from . more)
      (and (equal? t t2)
           `(((vector ,t) ,at (vector ,t) ,from ,@more) void))))

  (define-specialization $vector-copy! (_ 'crunch_vector_copy_into))

  (define-overloaded-primitive (scheme#vector-fill! ? ? #!optional integer
  	integer) void $vector-fill!
  	((('vector t) t2 more ...)
  	  (and (equal? t t2)
  	       `(((vector ,t) ,t ,@more) void))))

  (define-specialization $vector-fill!
    (('(vector integer) _ ...) 'crunch_fill_u8vector)
    (('(vector char) _ ...) 'crunch_fill_charvector)
    (('(vector float) _ ...) 'crunch_fill_f64vector)
    (('(vector complex) _ ...) 'crunch_fill_c128vector)
    (('(vector boolean) _ ...) 'crunch_fill_u8vector)
    (_ 'crunch_fill_vector))

  (define-overloaded-primitive (scheme#make-vector integer ?) ?
    $make-vector
    (('integer t) `((integer ,t) (vector ,t))))

  (define-specialization $make-vector
    ((_ 'integer) 'crunch_make_intvector)
    ((_ 'char) 'crunch_make_charvector)
    ((_ 'boolean) 'crunch_make_booleanvector)
    ((_ 'float) 'crunch_make_floatvector)
    ((_ 'complex) 'crunch_make_complexvector)
    (_ 'crunch_make_vector))

  (define-primitive (scheme#char=? char char) boolean $eq?)
  (define-primitive (scheme#char>? char char) boolean $>)
  (define-primitive (scheme#char<? char char) boolean $<)
  (define-primitive (scheme#char>=? char char) boolean $>=)
  (define-primitive (scheme#char<=? char char) boolean $<=)

  (define-primitive (chicken.bytevector#make-bytevector integer #!optional
    integer) (numvector u8) crunch_make_u8vector)

  (define-primitive (chicken.bytevector#bytevector-length (numvector u8)) integer
    crunch_vector_length)

  (define-primitive (chicken.bytevector#bytevector-u8-ref (numvector u8)
    integer) integer crunch_numvector_ref)

  (define-primitive (chicken.bytevector#bytevector-u8-set! (numvector u8)
    integer integer) void crunch_numvector_set)

  (define-primitive (chicken.bytevector#bytevector-copy (numvector u8)
    #!optional integer integer) (numvector u8) $bytevector-copy)

  (define-specialization $bytevector-copy
    ((_) 'crunch_u8vector_copy1)
    (_ 'crunch_u8vector_copy))

  (define-primitive (chicken.bytevector#bytevector-append (numvector u8) (numvector u8)
  	#!optional (numvector u8) (numvector u8) (numvector u8) (numvector u8)
  	(numvector u8) (numvector u8) (numvector u8) (numvector u8))
  	(numvector u8) crunch_u8vector_append)

  (define-primitive (chicken.bytevector#bytevector-copy! (numvector u8)
    integer (numvector u8) #!optional integer integer) void
    crunch_u8vector_copy_into)

  (define-inline-primitive (scheme#string-for-each proc str)
    (let ((loop (unspecified)))
      (begin
       (set! loop
         (lambda (proc i)
           (if (crunch.internal#%string-iterator-at-end? i)
               (unspecified)
               (begin
                 (proc (crunch.internal#%string-iterator-ref i))
                 (loop proc (crunch.internal#%string-iterator-next i))))))
       (loop proc (crunch.internal#%string-iterator str)))))

  (define-inline-primitive (scheme#string-map proc str)
    (let ((loop (unspecified))
          (nstr (unspeceified))
          (str1 str))
      (begin
        (set! nstr (crunch.internal#%string-buffer str1))
        (set! loop
          (lambda (proc i j)
            (if (crunch.internal#%string-iterator-at-end? i)
                (begin
                  (crunch.internal#%string-buffer-finalize nstr j)
                  nstr)
                (begin
                  (crunch.internal#%string-buffer-add nstr j
                    (proc (crunch.internal#%string-iterator-ref i)))
                  (loop proc (crunch.internal#%string-iterator-next i))))))
        (loop proc
              (crunch.internal#%string-iterator str)
              (crunch.internal#%string-iterator nstr)))))

  (define-primitive (scheme#string-ref string integer) char crunch_string_ref)

  (define-primitive (scheme#string-set! string integer char) void crunch_string_set)

  (define-primitive (scheme#substring string integer integer) string crunch_substring)

  (define-primitive (scheme#vector-append ? ? #!optional ? ? ? ? ? ? ? ?) ? $vector-append)

  (define-primitive (scheme#string-copy string #!optional integer integer)
    string crunch_string_copy)

  (define-primitive (scheme#string-copy! string integer string #!optional
    integer integer) void crunch_string_copy_into)

  (define-primitive (scheme#string-fill! string char #!optional integer
    integer) void crunch_string_fill)

  (define-specialization $vector-append
    ((('vector t) more ...)
      (and (every (match-lambda (('vector t2) (equal? t t2)) (_ #f)) more)
      	   (case t
      	     ((boolean) 'crunch_u8vector_append)
      	     ((integer) 'crunch_intvector_append)
      	     ((float) 'crunch_f64vector_append)
      	     ((complex) 'crunch_c128vector_append)
      	     (else 'crunch_vector_append))))
    ((('numvector t) more ...)
      (and (every (match-lambda (('numvector t2) (equal? t t2)) (_ #f)) more)
      	   (string->symbol (string-append "crunch_" (symbol->string t) "vector_append")))))

  (define-primitive (scheme#string-append string string
  				     #!optional string string string string string string string string) string
    crunch_string_append)

  (define-overloaded-primitive (scheme#max ? ?) ? $max
    (((and t1 (or 'integer 'float 'complex))
      (and t2 (or 'integer 'float 'complex)))
     (and (equal? t1 t2) `((,t1 ,t1) ,t1))))

  (define-specialization $max (_ 'crunch_max))

  (define-overloaded-primitive (scheme#min ? ?) ? $min
    (((and t1 (or 'integer 'float 'complex))
      (and t2 (or 'integer 'float 'complex)))
     (and (equal? t1 t2) `((,t1 ,t1) ,t1))))

  (define-specialization $min (_ 'crunch_min))

  (define-primitive (scheme#integer->char integer) char crunch_integer_to_char)
  (define-primitive (scheme#char->integer integer) char crunch_char_to_integer)

  (define-primitive (scheme#boolean=? boolean boolean) boolean $boolean=?)

  (define-specialization $boolean=? (_ '(operator ==)))

  (define-overloaded-primitive (scheme#inexact->exact ?) integer $exact
    (((and t (or 'integer 'float 'complex))) `((,t) integer)))

  (define-specialization $exact
    (('integer) '(operator + 0))
    (('float) 'crunch_exact_float)
    (('complex) 'crunch_exact_complex))

  (define-overloaded-primitive (scheme#exact->inexact ?) ? $inexact
    (('integer) '((integer) float))
    (('float) '((float) float))
    (('complex) '((complex) complex)))

  (define-specialization $inexact
    (('integer) 'crunch_inexact_integer)
    (('float) '(operator + 0))
    (('complex) '(operator + 0)))

  (define-primitive (scheme#string->vector string #!optional integer integer)
    (vector char) crunch_string_to_vector)

  (define-primitive (scheme#vector->string (vector char) #!optional integer integer)
    string crunch_vector_to_string)

  (define-primitive (scheme#floor float) float $floor)

  (define-specialization $floor (_ 'floor))

  (define-primitive (scheme#ceiling float) float $ceiling)

  (define-specialization $ceiling (_ 'ceil))

  (define-primitive (scheme#truncate float) float $truncate)

  (define-specialization $truncate (_ 'trunc))

  (define-primitive (scheme#round float) float $round)

  (define-specialization $round (_ 'round))

  (define-primitive (scheme#even? integer) boolean crunch_even)
  (define-primitive (scheme#odd? integer) boolean crunch_odd)

  (define-primitive (scheme#textual-port? port) boolean crunch_textual_port)
  (define-primitive (scheme#binary-port? port) boolean crunch_binary_port)
  (define-primitive (scheme#input-port? port) boolean crunch_input_port)
  (define-primitive (scheme#output-port? port) boolean crunch_output_port)

  (define-primitive (chicken.base#flush-output #!optional port) void crunch_flush_output_port)

  (define-overloaded-primitive (scheme#number->string ? #!optional integer) string
    $number->string
    (((and t (or 'integer 'float 'complex)) . more) `((,t ,@more) string)))

  (define-specialization $number->string
    (('integer . more) 'crunch_integer_to_string)
    (('float . more) 'crunch_float_to_string)
    (('complex . more) 'crunch_complex_to_string))

  (define-primitive (chicken.base#error string #!optional ? ? ?) noreturn
  	crunch_error)

  (define-primitive (scheme#current-input-port #!optional port) port
  	crunch_current_input_port)

  (define-primitive (scheme#current-output-port #!optional port) port
    crunch_current_output_port)

  (define-primitive (chicken.base#current-error-port #!optional port) port
    crunch_current_error_port)

  (define-primitive (scheme#close-port port) void crunch_close_port)
  (define-primitive (scheme#close-input-port port) void crunch_close_port)
  (define-primitive (scheme#close-output-port port) void crunch_close_port)

  (define-inline-primitive (scheme#call-with-port port proc)
    (let ((p port))
      (let ((r (proc p)))
        (begin (scheme#close-port p) r))))

  (define-primitive (scheme#open-input-string) port crunch_open_input_string)
  (define-primitive (scheme#open-input-bytevector) port crunch_open_input_bytevector)
  (define-primitive (scheme#open-output-string) port crunch_open_output_string)
  (define-primitive (scheme#open-output-bytevector) port crunch_open_output_bytevector)

  (define-primitive (scheme#get-output-string port) string crunch_get_output_string)
  (define-primitive (scheme#get-output-bytevector port) (numvector u8)
  	crunch_get_output_bytevector)

  (define-primitive (scheme#input-port-open? port) boolean crunch_port_open)
  (define-primitive (scheme#output-port-open? port) boolean crunch_port_open)

  (define-primitive (chicken.io#write-byte integer #!optional port)
    void crunch_write_u8)

  (define-primitive (scheme#write-bytevector (numvector u8) #!optional
    port integer integer) void crunch_write_bytevector)

  (define-primitive (scheme#write-string string #!optional port integer
    integer) void crunch_write_string)

  (define-primitive (scheme#write-char char #!optional port) void
    crunch_write_char)

  (define-primitive (scheme#peek-char #!optional port) char
    crunch_peek_char)

  (define-primitive (scheme#peek-u8 #!optional port) integer
    crunch_peek_u8)

  (define-primitive (scheme#char-ready? #!optional port) boolean
    crunch_char_ready)

  (define-primitive (scheme#u8-ready? #!optional port) boolean
    crunch_u8_ready)

  (define-primitive (chicken.io#read-byte #!optional port) integer
    crunch_read_u8)

  (define-primitive (chicken.io#read-string integer #!optional port)
    string crunch_read_string)

  (define-primitive (chicken.io#read-bytevector integer #!optional port)
    (numvector u8) crunch_read_bytevector)

  (define-primitive (scheme#read-char #!optional port) char
    crunch_read_char)

  (define-primitive (chicken.io#read-bytevector! (numvector u8)
    #!optional port integer integer) integer crunch_read_bytevector_into)

  (define-primitive (chicken.bytevector#utf8->string (numvector u8)
  	#!optional integer integer) string crunch_utf8_to_string)

  (define-primitive (chicken.bytevector#string->utf8 string
  	#!optional integer integer) (numvector u8) crunch_string_to_utf8)

  ;; (scheme inexact)

  (define-overloaded-primitive (scheme#acos ?) float $acos
    (args (inexact-function args)))

  (define-specialization $acos
    (((or 'integer 'float)) 'acos)
    (('complex) 'cacos))

  (define-overloaded-primitive (scheme#asin ?) float $asin
    (args (inexact-function args)))

  (define-specialization $asin
    (((or 'integer 'float)) 'asin)
    (('complex) 'casin))

  (define-overloaded-primitive (scheme#atan ? #!optional ?) float $atan
    (('complex) '((complex) complex))
    (((and t (or 'integer 'float))) `((,t) float))
    (((and t (or 'integer 'float)) (and t2 (or 'integer 'float))) `((,t ,t2) float)))

  (define-specialization $atan
    ((_ _) 'crunch_atan2)
    (((or 'integer 'float)) 'crunch_atan)
    (('complex) 'crunch_catan))

  (define-overloaded-primitive (scheme#sin ?) float $sin
    (args (inexact-function args)))

  (define-specialization $sin
    (((or 'integer 'float)) 'sin)
    (('complex) 'csin))

  (define-overloaded-primitive (scheme#cos ?) float $cos
    (args (inexact-function args)))

  (define-specialization $cos
    (((or 'integer 'float)) 'cos)
    (('complex) 'ccos))

  (define-overloaded-primitive (scheme#exp ?) float $exp
    (args (inexact-function args)))

  (define-specialization $exp
    (((or 'integer 'float)) 'exp)
    (('complex) 'cexp))

  (define-overloaded-primitive (scheme#log ?) float $log
    (args (inexact-function args)))

  (define-specialization $log
    (((or 'integer 'float)) 'log)
    (('complex) 'clog))

  (define-overloaded-primitive (scheme#tan ?) float $tan
    (args (inexact-function args)))

  (define-specialization $tan
    (((or 'integer 'float)) 'tan)
    (('complex) 'ctan))

  (define-overloaded-primitive (scheme#sqrt ?) float $sqrt
    (args (inexact-function args)))

  (define-specialization $sqrt
    (((or 'integer 'float)) 'sqrt)
    (('complex) 'csqrt))

  (define-overloaded-primitive (chicken.base#finite? ?) boolean $finite?
    (args (numeric-predicate args)))

  (define-specialization $finite?
    (((or 'integer 'float)) 'isfinite)
    (('complex) 'crunch_complex_finitep))

  (define-overloaded-primitive (chicken.base#infinite? ?) boolean $infinite?
    (args (numeric-predicate args)))

  (define-specialization $infinite?
    (((or 'integer 'float)) 'isinf)
    (('complex) 'crunch_complex_infinitep))

  (define-overloaded-primitive (chicken.base#nan? ?) boolean $finite?
    (args (numeric-predicate args)))

  (define-specialization $nan?
    (((or 'integer 'float)) 'isnan)
    (('complex) 'crunch_complex_nanp))

  ;; (scheme write)

  (define-overloaded-primitive (scheme.write#write ? #!optional port)
    void $write
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _) ('vector _) ('numvector _)))
      port)
      `((,t port) void))
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _) ('vector _) ('numvector _))))
      `((,t) void)))

  (define-specialization $write
    (((or 'pointer 'port) . _) 'crunch_write_pointer)
    (('string . _) 'crunch_write_string_readable)
    (('char . _) 'crunch_write_char_readable)
    (((and t (or 'boolean 'integer 'float 'complex)) . _)
      (string->symbol (string-append "crunch_write_" (symbol->string t))))
    (('(vector string) . _) 'crunch_write_stringvector)
    (('(vector integer) . _) 'crunch_write_intvector)
    (('(vector float) . _) 'crunch_write_f64vector)
    (('(vector char) . _) 'crunch_write_charvector)
    (('(vector complex) . _) 'crunch_write_c128vector)
    (('(vector boolean) . _) 'crunch_write_boolvector)
    ((('vector _) . _) 'crunch_write_ptrvector)
    (('(numvector u8) . _) 'crunch_write_u8vector)
    (('(numvector s8) . _) 'crunch_write_s8vector)
    (('(numvector u16) . _) 'crunch_write_u16vector)
    (('(numvector s16) . _) 'crunch_write_s16vector)
    (('(numvector u32) . _) 'crunch_write_u32vector)
    (('(numvector s32) . _) 'crunch_write_s32vector)
    (('(numvector u64) . _) 'crunch_write_u64vector)
    (('(numvector s64) . _) 'crunch_write_s64vector)
    (('(numvector f32) . _) 'crunch_write_f32vector)
    (('(numvector f64) . _) 'crunch_write_f64vector)
    (('(numvector c64) . _) 'crunch_write_c64vector)
    (('(numvector c128) . _) 'crunch_write_c128vector)
    ((((or 'struct 'union) _) . _) 'crunch_write_struct)
    (((or ('pointer _) ('procedure _ _)) . _) 'crunch_write_pointer))

  ;; write-simple
  (define-overloaded-primitive (scheme#write ? #!optional port)
    void $write-simple
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _) ('vector _) ('numvector _)))
      port)
      `((,t port) void))
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _) ('vector _) ('numvector _))))
      `((,t) void)))

  (define-specialization $write-simple = $write)

  (define-overloaded-primitive (scheme.write#display ? #!optional port)
    void $display
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _)))
      port)
      `((,t port) void))
    (((and t (or 'boolean 'integer 'float 'complex 'char 'string 'pointer 'port
                 ('pointer _) ('procedure _ _))))
      `((,t) void)))

  (define-specialization $display
    (((or 'pointer 'port) . _) 'crunch_write_pointer)
    (((and t (or 'char 'string)) . _)
      (string->symbol (string-append "crunch_display_" (symbol->string t))))
    (((and t (or 'boolean 'integer 'float 'complex)) . _)
      (string->symbol (string-append "crunch_write_" (symbol->string t))))
    ('(vector string) 'crunch_display_stringvector)
    ('(vector integer) 'crunch_write_intvector)
    ('(vector char) 'crunch_write_charvector)
    ('(vector float) 'crunch_write_f64vector)
    ('(vector complex) 'crunch_write_c128vector)
    ('(vector boolean) 'crunch_write_boolvector)
    (('vector _) 'crunch_write_ptrvector)
    ('(numvector u8) 'crunch_write_u8vector)
    ('(numvector s8) 'crunch_write_s8vector)
    ('(numvector u16) 'crunch_write_u16vector)
    ('(numvector s16) 'crunch_write_s16vector)
    ('(numvector u32) 'crunch_write_u32vector)
    ('(numvector s32) 'crunch_write_s32vector)
    ('(numvector u64) 'crunch_write_u64vector)
    ('(numvector s64) 'crunch_write_s64vector)
    ('(numvector f32) 'crunch_write_f32vector)
    ('(numvector f64) 'crunch_write_f64vector)
    ('(numvector c64) 'crunch_write_c64vector)
    ('(numvector c128) 'crunch_write_c128vector)
    ((((or 'struct 'union) _) . _) 'crunch_write_struct)
    (((or ('pointer _) ('procedure _ _)) . _) 'crunch_write_pointer))

  ;; (scheme complex)

  (define-overloaded-primitive (scheme#make-rectangular ? ?)
    complex $make-rectangular
    (((and t1 (or 'integer 'float)) (and t2 (or 'integer 'float)))
      `((,t1 ,t2) complex)))

  (define-specialization $make-rectangular (_ 'crunch_make_rectangular))

  (define-overloaded-primitive (scheme#make-polar ? ?)
    complex $make-polar
    (((and t1 (or 'integer 'float)) (and t2 (or 'integer 'float)))
      `((,t1 ,t2) complex)))

  (define-specialization $make-polar (_ 'crunch_make_polar))

  (define-overloaded-primitive (scheme#real-part ?) ? $real-part
    (((and t (or 'integer 'float))) `((,t) ,t))
    (('complex) '((complex) float)))

  (define-specialization $real-part
    (((or 'integer 'float)) '(operator + 0))
    (('complex) 'creal))

  (define-overloaded-primitive (scheme#imag-part ?) ? $imag-part
    (((and t (or 'integer 'float))) `((,t) ,t))
    (('complex) '((complex) float)))

  (define-specialization $imag-part
    (((or 'integer 'float)) '(operator + 0))
    (('complex) 'cimag))

  (define-overloaded-primitive (scheme#magnitude ?) ? $magnitude
    (((and t (or 'integer 'float))) `((,t) ,t))
    (('complex) '((complex) float)))

  (define-specialization $magnitude
    (((or 'integer 'float)) 'abs)
    (('complex) 'cabs))

  (define-overloaded-primitive (scheme#angle ?) ? $angle
    (((and t (or 'integer 'float))) `((,t) ,t))
    (('complex) '((complex) float)))

  (define-specialization $angle
    (((or 'integer 'float)) 'crunch_float_angle)
    (('complex) 'carg))

  ;; (chicken bitwise)

  (define-primitive (chicken.bitwise#bitwise-and integer integer) integer $bitwise-and)

  (define-specialization $bitwise-and (_ '(operator &)))

  (define-primitive (chicken.bitwise#bitwise-ior integer integer) integer $bitwise-ior)

  (define-specialization $bitwise-ior (_ '(operator |\||)))

  (define-primitive (chicken.bitwise#bitwise-xor integer integer) integer $bitwise-xor)

  (define-specialization $bitwise-xor (_ '(operator ^)))

  (define-primitive (chicken.bitwise#bitwise-not integer) integer $bitwise-not)

  (define-specialization $bitwise-not (_ '(operator ^ -1)))

  (define-primitive (chicken.bitwise#arithmetic-shift integer integer) integer
	$arithmetic-shift)

  (define-specialization $arithmetic-shift (_ 'crunch_arithmetic_shift))

  (define-primitive (chicken.bitwise#integer-length integer) integer $integer-length)

  (define-specialization $integer-length (_ 'crunch_integer_length))

  ;; (chicken base)

  (define-overloaded-primitive (chicken.base#add1 ?) ? $add1
    (args (numeric-function args)))

  (define-specialization $add1 (_ '(operator + 1)))

  (define-overloaded-primitive (chicken.base#sub1 ?) ? $sub1
    (args (numeric-function args)))

  (define-specialization $sub1 (_ '(operator - 1)))

  (define-overloaded-primitive (chicken.base#void #!optional ? ? ? ? ? ? ? ? ? ?)
    void $void
    (args (list args 'void)))

  (define-specialization $void (_ 'crunch_void))

  (define-custom-inline-primitive scheme#apply
    (lambda (n)
      (let* ((loc (node-location n))
             (ns (cdr (node-subnodes n)))
             (vn (last ns))
             (vars (map (lambda _ (gensym 'v)) ns)))
        (node*
          loc 'lambda (list (gensym 'f) (length ns) vars vars #f)
          (apply node*
            loc 'call (list #t)
            (append
              (butlast ns)
              (list (node*
                      loc 'call (list #f)
                      (node* loc 'global-ref
                        (list 'crunch.internal#value-ref #f 'value-ref))
                      vn))))))))

  ;; (scheme char)

  (define-primitive (scheme#char-alphabetic? char) boolean
  	crunch_utf_isalpha)

  (define-primitive (scheme#char-whitespace? char) boolean
  	crunch_utf_isspace)

  (define-primitive (scheme#char-upper-case? char) boolean
  	crunch_utf_isupper)

  (define-primitive (scheme#char-lower-case? char) boolean
  	crunch_utf_islower)

  (define-primitive (scheme#char-numeric? char) boolean
  	crunch_utf_isdigit)

  (define-primitive (scheme#char-downcase char) char
    crunch_utf_char_downcase)

  (define-primitive (scheme#char-upcase char) char
    crunch_utf_char_upcase)

  (define-primitive (scheme#char-foldcase char) char
    crunch_utf_char_foldcase)

  (define-primitive (scheme#char-ci=? char char) boolean
    crunch_char_ci_equal)

  (define-primitive (scheme#char-ci>? char char) boolean
    crunch_char_ci_greater)

  (define-primitive (scheme#char-ci<? char char) boolean
    crunch_char_ci_less)

  (define-primitive (scheme#char-ci>=? char char) boolean
    crunch_char_ci_greater_or_equal)

  (define-primitive (scheme#char-ci<=? char char) boolean
    crunch_char_ci_less_or_equal)

  (define-primitive (scheme#string-ci=? string string) boolean crunch_string_equal_ci)

  (define-primitive (scheme#string-ci>? string string) boolean crunch_string_greater_ci)

  (define-primitive (scheme#string-ci<? string string) boolean crunch_string_less_ci)

  (define-primitive (scheme#string-ci>=? string string) boolean crunch_string_greater_or_equal_ci)

  (define-primitive (scheme#string-ci<=? string string) boolean crunch_string_less_or_equal_ci)

  (define-primitive (scheme#string-upcase string) string crunch_string_upcase)

  (define-primitive (scheme#string-downcase string) string crunch_string_downcase)

  (define-primitive (scheme#string-foldcase string) string crunch_string_foldcase)

  ;; (scheme file)

  (define-primitive (scheme#open-input-file string) port crunch_open_input_file)
  (define-primitive (scheme#open-output-file string) port crunch_open_output_file)

  (define-primitive (scheme.file#open-binary-input-file string) port
  	crunch_open_binary_input_file)
  (define-primitive (scheme.file#open-binary-output-file string) port
  	crunch_open_binary_output_file)

  (define-inline-primitive (scheme#call-with-input-file fname proc)
    (scheme#call-with-port
      (scheme#open-input-file fname)
      proc))

  (define-inline-primitive (scheme#call-with-output-file fname proc)
    (scheme#call-with-port
      (scheme#open-output-file fname)
      proc))

  (define-inline-primitive (scheme#with-input-from-file fname proc)
    (let ((p (scheme#current-input-port)))
      (scheme#current-input-port (scheme#open-input-file fname))
      (let ((r (proc)))
        (begin (scheme#current-input-port p) r))))

  (define-inline-primitive (scheme#with-output-from-file fname proc)
    (let ((p (scheme#current-output-port)))
      (scheme#current-output-port (scheme#open-input-file fname))
      (let ((r (proc)))
        (begin (scheme#current-output-port p) r))))

  (define-primitive (chicken.file#delete-file string) void crunch_delete_file)

  (define-primitive (scheme.file#file-exists? string) boolean
    crunch_file_exists)

  ;; (scheme process-context)

  (define-overloaded-primitive (scheme.process-context#exit #!optional ?) noreturn $exit
    (() '(() noreturn))
    (('boolean) '((boolean) noreturn))
    (('integer) '((integer) noreturn)))

  (define-specialization $exit (_ 'crunch_exit))

  (define-overloaded-primitive (chicken.base#emergency-exit #!optional ?) noreturn $emergency-exit
    (() '(() noreturn))
    (('boolean) '((boolean) noreturn))
    (('integer) '((integer) noreturn)))

  (define-specialization $emergency-exit (_ 'crunch_emergency_exit))

  ;; (scheme time)

  (define-primitive (scheme.time#current-jiffy) integer crunch_current_jiffy)

  (define-primitive (scheme.time#current-second) integer
    crunch_current_second)

  (define-primitive (scheme.time#jiffies-per-second) integer
    crunch_jiffies_per_second)

  ;; (chicken fixnum)

  (define-primitive (chicken.fixnum#fx+ integer integer) integer $fx+)
  (define-primitive (chicken.fixnum#fx- integer integer) integer $fx-)
  (define-primitive (chicken.fixnum#fx* integer integer) integer $fx*)
  (define-primitive (chicken.fixnum#fx/ integer integer) integer $fx/)
  (define-primitive (chicken.fixnum#fxmax integer integer) integer crunch_max)
  (define-primitive (chicken.fixnum#fxmin integer integer) integer crunch_min)
  (define-primitive (chicken.fixnum#fx= integer integer) boolean $fx=)
  (define-primitive (chicken.fixnum#fx> integer integer) boolean $fx>)
  (define-primitive (chicken.fixnum#fx< integer integer) boolean $fx<)
  (define-primitive (chicken.fixnum#fx>= integer integer) boolean $fx>=)
  (define-primitive (chicken.fixnum#fx<= integer integer) boolean $fx<=)
  (define-primitive (chicken.fixnum#fxand integer integer) integer $fxand)
  (define-primitive (chicken.fixnum#fxior integer integer) integer $fxior)
  (define-primitive (chicken.fixnum#fxxor integer integer) integer $fxxor)
  (define-primitive (chicken.fixnum#fxmod integer integer) integer $fxmod)
  (define-primitive (chicken.fixnum#fxrem integer integer) integer crunch_fxrem)
  (define-primitive (chicken.fixnum#fxnot integer) integer $fxnot)
  (define-primitive (chicken.fixnum#fxneg integer) integer $fxneg)
  (define-primitive (chicken.fixnum#fxodd? integer) boolean crunch_odd)
  (define-primitive (chicken.fixnum#fxeven? integer) boolean crunch_even)
  (define-primitive (chicken.fixnum#fxshr integer integer) integer $fxshr)
  (define-primitive (chicken.fixnum#fxshl integer integer) integer $fxshl)
  (define-primitive (chicken.fixnum#fxlen integer integer) integer crunch_fxlen)

  (define-specialization $fx+ (_ '(operator +)))
  (define-specialization $fx- (_ '(operator -)))
  (define-specialization $fx* (_ '(operator *)))
  (define-specialization $fx/ (_ '(operator /)))
  (define-specialization $fxand (_ '(operator &)))
  (define-specialization $fxior (_ '(operator |\||)))
  (define-specialization $fxxor (_ '(operator ^)))
  (define-specialization $fxmod (_ '(operator %)))
  (define-specialization $fx= (_ '(operator ==)))
  (define-specialization $fx< (_ '(operator <)))
  (define-specialization $fx> (_ '(operator >)))
  (define-specialization $fx<= (_ '(operator <=)))
  (define-specialization $fx>= (_ '(operator >=)))
  (define-specialization $fxnot (_ '(operator ^ -1)))
  (define-specialization $fxneg (_ '(operator * -1)))
  (define-specialization $fxshr (_ '(operator >>)))
  (define-specialization $fxshl (_ '(operator <<)))

  ;; (chicken flonum)

  (define-primitive (chicken-flonum#fp+ float float) float $fp+)
  (define-primitive (chicken-flonum#fp- float float) float $fp-)
  (define-primitive (chicken-flonum#fp* float float) float $fp*)
  (define-primitive (chicken-flonum#fp/ float float) float $fp/)
  (define-primitive (chicken-flonum#fp*+ float float float) float fma)
  (define-primitive (chicken-flonum#fp= float float) boolean $fp=)
  (define-primitive (chicken-flonum#fp> float float) boolean $fp>)
  (define-primitive (chicken-flonum#fp< float float) boolean $fp<)
  (define-primitive (chicken-flonum#fp>= float float) boolean $fp>=)
  (define-primitive (chicken-flonum#fp<= float float) boolean $fp<=)
  (define-primitive (chicken-flonum#fpabs float) float fabs)
  (define-primitive (chicken-flonum#fpsin float) float sin)
  (define-primitive (chicken-flonum#fpcos float) float cos)
  (define-primitive (chicken-flonum#fptan float) float tan)
  (define-primitive (chicken-flonum#fpsinh float) float sinh)
  (define-primitive (chicken-flonum#fpcosh float) float cosh)
  (define-primitive (chicken-flonum#fptanh float) float tanh)
  (define-primitive (chicken-flonum#fpasin float) float asin)
  (define-primitive (chicken-flonum#fpacos float) float acos)
  (define-primitive (chicken-flonum#fpatan float) float atan)
  (define-primitive (chicken-flonum#fpasinh float) float asinh)
  (define-primitive (chicken-flonum#fpacosh float) float acosh)
  (define-primitive (chicken-flonum#fpatanh float) float atanh)
  (define-primitive (chicken-flonum#fpatan2 float float) float atan2)
  (define-primitive (chicken-flonum#fpceiling float) float ceil)
  (define-primitive (chicken-flonum#fpfloor float) float floor)
  (define-primitive (chicken-flonum#fpround float) float round)
  (define-primitive (chicken-flonum#fptruncate float) float trunc)
  (define-primitive (chicken-flonum#fpexp float) float exp)
  (define-primitive (chicken-flonum#fpexpt float float) float pow)
  (define-primitive (chicken-flonum#fplog float) float log)
  (define-primitive (chicken-flonum#fpsqrt float) float sqrt)
  (define-primitive (chicken-flonum#fpneg float) float $fpneg)
  (define-primitive (chicken-flonum#fpmax float float) float crunch_max)
  (define-primitive (chicken-flonum#fpmin float float) float crunch_min)

  (define-specialization $fp+ (_ '(operator +)))
  (define-specialization $fp- (_ '(operator -)))
  (define-specialization $fp* (_ '(operator *)))
  (define-specialization $fp/ (_ '(operator /)))
  (define-specialization $fp= (_ '(operator ==)))
  (define-specialization $fp> (_ '(operator >)))
  (define-specialization $fp< (_ '(operator <)))
  (define-specialization $fp>= (_ '(operator >=)))
  (define-specialization $fp<= (_ '(operator <=)))
  (define-specialization $fpneg (_ '(operator * -1)))

  ;; (chicken memory)

  (define-primitive (chicken.memory#allocate integer) pointer crunch_allocate)

  (define-overloaded-primitive (chicken.memory#free ?) void $free
    (((and t (or 'pointer ('pointer _)))) `((,t) void)))

  (define-specialization $free (_ 'free))

  (define-primitive (chicken.memory#address->pointer integer) pointer crunch_address_to_pointer)
  (define-primitive (chicken.memory#pointer->address pointer) integer crunch_pointer_to_address)
  (define-primitive (chicken.memory#pointer=? pointer pointer) boolean $pointer=?)

  (define-specialization $pointer=? (_ '(operator ==)))

  (define-primitive (chicken.memory#pointer-u8-ref pointer) integer crunch_pointer_u8_ref)
  (define-primitive (chicken.memory#pointer-s8-ref pointer) integer crunch_pointer_s8_ref)
  (define-primitive (chicken.memory#pointer-u16-ref pointer) integer crunch_pointer_u16_ref)
  (define-primitive (chicken.memory#pointer-s16-ref pointer) integer crunch_pointer_s16_ref)
  (define-primitive (chicken.memory#pointer-u32-ref pointer) integer crunch_pointer_u32_ref)
  (define-primitive (chicken.memory#pointer-s32-ref pointer) integer crunch_pointer_s32_ref)
  (define-primitive (chicken.memory#pointer-u64-ref pointer) integer crunch_pointer_u64_ref)
  (define-primitive (chicken.memory#pointer-s64-ref pointer) integer crunch_pointer_s64_ref)
  (define-primitive (chicken.memory#pointer-f32-ref pointer) float crunch_pointer_f32_ref)
  (define-primitive (chicken.memory#pointer-f64-ref pointer) float crunch_pointer_f64_ref)
  (define-primitive (chicken.memory#pointer-c64-ref pointer) complex crunch_pointer_c64_ref)
  (define-primitive (chicken.memory#pointer-c128-ref pointer) complex crunch_pointer_c128_ref)

  (define-primitive (chicken.memory#pointer-u8-set! pointer integer) void crunch_pointer_u8_set)
  (define-primitive (chicken.memory#pointer-s8-set! pointer integer) void crunch_pointer_s8_set)
  (define-primitive (chicken.memory#pointer-u16-set! pointer integer) void crunch_pointer_u16_set)
  (define-primitive (chicken.memory#pointer-s16-set! pointer integer) void crunch_pointer_s16_set)
  (define-primitive (chicken.memory#pointer-u32-set! pointer integer) void crunch_pointer_u32_set)
  (define-primitive (chicken.memory#pointer-s32-set! pointer integer) void crunch_pointer_s32_set)
  (define-primitive (chicken.memory#pointer-u64-set! pointer integer) void crunch_pointer_u64_set)
  (define-primitive (chicken.memory#pointer-s64-set! pointer integer) void crunch_pointer_s64_set)
  (define-primitive (chicken.memory#pointer-f32-set! pointer float) void crunch_pointer_f32_set)
  (define-primitive (chicken.memory#pointer-f64-set! pointer float) void crunch_pointer_f64_set)
  (define-primitive (chicken.memory#pointer-c64-set! pointer complex) void crunch_pointer_c64_set)
  (define-primitive (chicken.memory#pointer-c128-set! pointer complex) void crunch_pointer_c128_set)

  ;; (chicken number-vector)

  (define-primitive (chicken.number-vector#u8vector-ref (numvector u8) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#s8vector-ref (numvector s8) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#u16vector-ref (numvector u16) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#s16vector-ref (numvector s16) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#u32vector-ref (numvector u32) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#s32vector-ref (numvector s32) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#u64vector-ref (numvector u64) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#s64vector-ref (numvector s64) integer) integer
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#f32vector-ref (numvector f32) integer) float
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#f64vector-ref (numvector f64) integer) float
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#c64vector-ref (numvector c64) integer) complex
    crunch_numvector_ref)
  (define-primitive (chicken.number-vector#c128vector-ref (numvector c128) integer) complex
    crunch_numvector_ref)

  (define-primitive (chicken.number-vector#u8vector-set! (numvector u8) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#s8vector-set! (numvector s8) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#u16vector-set! (numvector u16) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#s16vector-set! (numvector s16) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#u32vector-set! (numvector u32) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#s32vector-set! (numvector s32) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#u64vector-set! (numvector u64) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#s64vector-set! (numvector s64) integer integer) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#f32vector-set! (numvector f32) integer float) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#f64vector-set! (numvector f64) integer float) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#c64vector-set! (numvector c64) integer complex) void
    crunch_numvector_set)
  (define-primitive (chicken.number-vector#c128vector-set! (numvector c128) integer complex) void
    crunch_numvector_set)

  (define-primitive (chicken.number-vector#make-u8vector integer #!optional integer) (numvector u8)
    crunch_make_u8vector)
  (define-primitive (chicken.number-vector#make-s8vector integer #!optional integer) (numvector s8)
    crunch_make_s8vector)
  (define-primitive (chicken.number-vector#make-u16vector integer #!optional integer) (numvector u16)
    crunch_make_u16vector)
  (define-primitive (chicken.number-vector#make-s16vector integer #!optional integer) (numvector s16)
    crunch_make_s16vector)
  (define-primitive (chicken.number-vector#make-u32vector integer #!optional integer) (numvector u32)
    crunch_make_u32vector)
  (define-primitive (chicken.number-vector#make-s32vector integer #!optional integer) (numvector s32)
    crunch_make_s32vector)
  (define-primitive (chicken.number-vector#make-u64vector integer #!optional integer) (numvector u64)
    crunch_make_u64vector)
  (define-primitive (chicken.number-vector#make-s64vector integer #!optional integer) (numvector s64)
    crunch_make_s64vector)
  (define-primitive (chicken.number-vector#make-f32vector integer #!optional integer) (numvector f32)
    crunch_make_f32vector)
  (define-primitive (chicken.number-vector#make-f64vector integer #!optional integer) (numvector f64)
    crunch_make_f64vector)
  (define-primitive (chicken.number-vector#make-c64vector integer #!optional integer) (numvector c64)
    crunch_make_c64vector)
  (define-primitive (chicken.number-vector#make-c128vector integer #!optional integer) (numvector c128)
    crunch_make_c128vector)

  (define-primitive (chicken.number-vector#u8vector-length (numvector u8)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#s8vector-length (numvector s8)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#u16vector-length (numvector u16)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#s16vector-length (numvector s16)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#u32vector-length (numvector u32)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#s32vector-length (numvector s32)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#u64vector-length (numvector u64)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#s64vector-length (numvector s64)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#f32vector-length (numvector f32)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#f64vector-length (numvector f64)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#c64vector-length (numvector c64)) integer
    crunch_vector_length)
  (define-primitive (chicken.number-vector#c128vector-length (numvector c128)) integer
    crunch_vector_length)

  (define-primitive (chicken.number-vector#subu8vector (numvector u8) integer integer)
    (numvector u8) crunch_subu8vector)
  (define-primitive (chicken.number-vector#subs8vector (numvector s8) integer integer)
    (numvector s8) crunch_subs8vector)
  (define-primitive (chicken.number-vector#subu16vector (numvector u16) integer integer)
    (numvector u16) crunch_subu16vector)
  (define-primitive (chicken.number-vector#subs16vector (numvector s16) integer integer)
    (numvector s16) crunch_subs16vector)
  (define-primitive (chicken.number-vector#subu32vector (numvector u32) integer integer)
    (numvector u32) crunch_subu32vector)
  (define-primitive (chicken.number-vector#subs32vector (numvector s32) integer integer)
    (numvector s32) crunch_subs32vector)
  (define-primitive (chicken.number-vector#subu64vector (numvector u64) integer integer)
    (numvector u64) crunch_subu64vector)
  (define-primitive (chicken.number-vector#subs64vector (numvector s64) integer integer)
    (numvector s64) crunch_subs64vector)
  (define-primitive (chicken.number-vector#subf32vector (numvector f32) integer integer)
    (numvector f32) crunch_subf32vector)
  (define-primitive (chicken.number-vector#subf64vector (numvector f64) integer integer)
    (numvector f64) crunch_subf64vector)
  (define-primitive (chicken.number-vector#subc64vector (numvector c64) integer integer)
    (numvector c64) crunch_subc64vector)
  (define-primitive (chicken.number-vector#subc128vector (numvector c128) integer integer)
    (numvector c128) crunch_subc128vector)

  ;; (crunch process-context)

  (define-primitive (crunch.process-context#command-line) (vector string)
    crunch_command_line)

  (define-primitive (crunch.process-context#get-environment-variable string
  	#!optional string) string crunch_get_environment_variable)

  ;; (crunch memory)

  (define-overloaded-primitive (crunch.memory#pointer-ref ?) ? $pointer-ref
    ((('pointer t)) `(((pointer ,t)) ,t)))

  (define-specialization $pointer-ref (_ `crunch_pointer_ref))

  (define-overloaded-primitive (crunch.memory#pointer-set! ? ?) void $pointer-set!
    ((('pointer t) t2) (and (equal? t t2) `(((pointer ,t) ,t) void))))

  (define-specialization $pointer-set! (_ `crunch_pointer_set))

  (define-overloaded-primitive (crunch.memory#pointer-index-ref ? integer) ? $pointer-index-ref
    ((('pointer t) _) `(((pointer ,t) integer) ,t)))

  (define-specialization $pointer-index-ref (_ `crunch_pointer_index_ref))

  (define-overloaded-primitive (crunch.memory#pointer-index-set! ? integer ?) void
    $pointer-index-set!
    ((('pointer t) _ t2) (and (equal? t t2) `(((pointer ,t) integer ,t) void))))

  (define-specialization $pointer-index-set! (_ `crunch_pointer_index-set))

  (define-primitive (crunch.memory#pointer-char-ref (pointer char)) char crunch_pointer_char_ref)
  (define-primitive (crunch.memory#pointer-char-set! (pointer char) char) void
  	crunch_pointer_char_set)

  (define-primitive (crunch.memory#pointer->string (pointer char)) string crunch_string_at)

  (define-primitive (crunch.memory#string->pointer string) (pointer char) crunch_data)

  ;; (crunch internal) - pseudo module used for intermediate compilation intrinsics

  (define-primitive (crunch.internal#%string-iterator string) pointer crunch_string_iterator)

  (define-primitive (crunch.internal#%string-iterator-at-end? pointer) boolean
  	crunch_string_iterator_at_end)

  (define-primitive (crunch.internal#%string-iterator-ref pointer) char
    crunch_string_iterator_ref)

  (define-primitive (crunch.internal#%string-iterator-next pointer) pointer
    crunch_string_iterator_next)

  (define-primitive (crunch.internal#%string-buffer string) string crunch_string_buffer)

  (define-primitive (crunch.internal#%string-buffer-add string pointer char) pointer
    crunch_string_buffer_add)

  (define-primitive (crunch.internal#%string-buffer-finalize string pointer) void
    crunch_string_buffer_finalize)

  (define-primitive (crunch.internal#%optional integer) boolean
    crunch_optional)

  (define-overloaded-primitive (crunch.internal#value-ref ?) ? $value-ref
    ((('values t)) `(((values ,t)) ,t)))

  (define-specialization $value-ref (_ 'crunch_identity))

)